package dev.fyloz.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import dev.fyloz.colorrecipesexplorer.exception.RestException
import dev.fyloz.colorrecipesexplorer.utils.months
import org.springframework.http.HttpStatus
import org.springframework.web.multipart.MultipartFile
import java.time.LocalDateTime
import javax.persistence.Column
import javax.persistence.Entity
import javax.persistence.Id
import javax.persistence.Table
import javax.validation.constraints.NotBlank

sealed class ConfigurationBase(
    @JsonIgnore
    val type: ConfigurationType,
    val lastUpdated: LocalDateTime
) {
    val key = type.key
    val requireRestart = type.requireRestart
    val editable = !type.computed
}

class Configuration(type: ConfigurationType, val content: String, lastUpdated: LocalDateTime) :
    ConfigurationBase(type, lastUpdated) {
    fun toEntity() =
        ConfigurationEntity(key, content, lastUpdated)
}

class SecureConfiguration(type: ConfigurationType, lastUpdated: LocalDateTime) :
    ConfigurationBase(type, lastUpdated)

@Entity
@Table(name = "configuration")
data class ConfigurationEntity(
    @Id
    @Column(name = "config_key")
    val key: String,

    val content: String,

    @Column(name = "last_updated")
    val lastUpdated: LocalDateTime
) {
    fun toConfiguration() =
        configuration(key.toConfigurationType(), content, lastUpdated)

    override fun equals(other: Any?) =
        other is ConfigurationEntity && key == other.key && content == other.content

    override fun hashCode(): Int {
        var result = key.hashCode()
        result = 31 * result + content.hashCode()
        return result
    }
}

data class ConfigurationDto(
    val key: String,

    @NotBlank
    val content: String
)

data class ConfigurationImageDto(
    val key: String,

    val image: MultipartFile
)

fun configuration(
    type: ConfigurationType,
    content: String = type.defaultContent.toString(),
    lastUpdated: LocalDateTime? = null
) = Configuration(type, content, lastUpdated ?: LocalDateTime.now())

fun configuration(
    dto: ConfigurationDto
) = with(dto) {
    configuration(type = key.toConfigurationType(), content = content)
}

fun secureConfiguration(
    type: ConfigurationType,
    lastUpdated: LocalDateTime? = null
) = SecureConfiguration(type, lastUpdated ?: LocalDateTime.now())

fun secureConfiguration(
    configuration: Configuration
) = secureConfiguration(configuration.type, configuration.lastUpdated)

enum class ConfigurationType(
    val key: String,
    val defaultContent: Any? = null,
    val computed: Boolean = false,
    val file: Boolean = false,
    val requireRestart: Boolean = false,
    val public: Boolean = false,
    val secure: Boolean = false
) {
    INSTANCE_NAME("instance.name", defaultContent = "Color Recipes Explorer", public = true),
    INSTANCE_LOGO_SET("instance.logo.set", defaultContent = false, public = true),
    INSTANCE_ICON_SET("instance.icon.set", defaultContent = false, public = true),
    INSTANCE_URL("instance.url", "http://localhost:9090", public = true),

    DATABASE_URL("database.url", defaultContent = "mysql://localhost/cre", file = true, requireRestart = true),
    DATABASE_USER("database.user", defaultContent = "cre", file = true, requireRestart = true),
    DATABASE_PASSWORD("database.password", defaultContent = "asecurepassword", file = true, requireRestart = true, secure = true),
    DATABASE_SUPPORTED_VERSION("database.version.supported", computed = true),

    RECIPE_APPROBATION_EXPIRATION("recipe.approbation.expiration", defaultContent = 4.months),

    TOUCH_UP_KIT_CACHE_PDF("touchupkit.pdf.cache", defaultContent = true),
    TOUCH_UP_KIT_EXPIRATION("touchupkit.expiration", defaultContent = 1.months),

    EMERGENCY_MODE_ENABLED("env.emergency", computed = true, public = true),
    BUILD_VERSION("env.build.version", computed = true),
    BUILD_TIME("env.build.time", computed = true),
    JAVA_VERSION("env.java.version", computed = true),
    OPERATING_SYSTEM("env.os", computed = true),

    GENERATED_ENCRYPTION_SALT("security.salt", file = true, requireRestart = true)
    ;

    override fun toString() = key
}

fun String.toConfigurationType() =
    ConfigurationType.values().firstOrNull { it.key == this }
        ?: throw InvalidConfigurationKeyException(this)

class InvalidConfigurationKeyException(val key: String) :
    RestException(
        "invalid-configuration-key",
        "Invalid configuration key",
        HttpStatus.BAD_REQUEST,
        "The configuration key '$key' does not exists",
        mapOf(
            "key" to key
        )
    )

class InvalidImageConfigurationException(val type: ConfigurationType) :
    RestException(
        "invalid-configuration-image",
        "Invalid image configuration",
        HttpStatus.BAD_REQUEST,
        "The configuration with the key '${type.key}' does not accept images as content",
        mapOf(
            "key" to type.key
        )
    )

class ConfigurationNotSetException(val type: ConfigurationType) :
    RestException(
        "unset-configuration",
        "Unset configuration",
        HttpStatus.NOT_FOUND,
        "The configuration with the key '${type.key}' is not set",
        mapOf(
            "key" to type.key
        )
    )

class CannotSetComputedConfigurationException(val type: ConfigurationType) :
    RestException(
        "cannot-set-computed-configuration",
        "Cannot set computed configuration",
        HttpStatus.BAD_REQUEST,
        "The configuration with the key '${type.key}' is a computed configuration and cannot be modified",
        mapOf(
            "key" to type.key
        )
    )
