package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.model.account.Group
import dev.fyloz.colorrecipesexplorer.model.validation.or
import dev.fyloz.colorrecipesexplorer.repository.RecipeRepository
import dev.fyloz.colorrecipesexplorer.service.config.ConfigurationService
import dev.fyloz.colorrecipesexplorer.service.files.WriteableFileService
import dev.fyloz.colorrecipesexplorer.utils.setAll
import org.springframework.context.annotation.Lazy
import org.springframework.context.annotation.Profile
import org.springframework.stereotype.Service
import org.springframework.web.multipart.MultipartFile
import java.io.File
import java.time.LocalDate
import java.time.Period
import javax.transaction.Transactional

interface RecipeService :
    ExternalModelService<Recipe, RecipeSaveDto, RecipeUpdateDto, RecipeOutputDto, RecipeRepository> {
    /** Checks if one or more recipes have the given [company]. */
    fun existsByCompany(company: Company): Boolean

    /** Checks if a recipe exists with the given [name] and [company]. */
    fun existsByNameAndCompany(name: String, company: Company): Boolean

    /** Checks if the approbation date of the given [recipe] is expired. */
    fun isApprobationExpired(recipe: Recipe): Boolean?

    /** Gets all recipes with the given [name]. */
    fun getAllByName(name: String): Collection<Recipe>

    /** Gets all recipes with the given [company]. */
    fun getAllByCompany(company: Company): Collection<Recipe>

    /** Updates the public data of a recipe with the given [publicDataDto]. */
    fun updatePublicData(publicDataDto: RecipePublicDataDto)

    /** Adds the given [mix] to the given [recipe]. */
    fun addMix(recipe: Recipe, mix: Mix): Recipe

    /** Removes the given [mix] from its recipe. */
    fun removeMix(mix: Mix): Recipe
}

@Service
@Profile("!emergency")
class RecipeServiceImpl(
    recipeRepository: RecipeRepository,
    val companyService: CompanyService,
    val mixService: MixService,
    val recipeStepService: RecipeStepService,
    @Lazy val groupService: GroupService,
    val recipeImageService: RecipeImageService,
    val configService: ConfigurationService
) :
    AbstractExternalModelService<Recipe, RecipeSaveDto, RecipeUpdateDto, RecipeOutputDto, RecipeRepository>(
        recipeRepository
    ),
    RecipeService {
    override fun idNotFoundException(id: Long) = recipeIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = recipeIdAlreadyExistsException(id)

    override fun Recipe.toOutput() = RecipeOutputDto(
        this.id!!,
        this.name,
        this.description,
        this.color,
        this.gloss,
        this.sample,
        this.approbationDate,
        isApprobationExpired(this),
        this.remark,
        this.company,
        this.mixes.map {
            with(mixService) {
                it.toOutput()
            }
        }.toSet(),
        this.groupsInformation,
        recipeImageService.getAllImages(this)
            .map { this.imageUrl(configService.getContent(ConfigurationType.INSTANCE_URL), it) }
            .toSet()
    )

    override fun existsByCompany(company: Company): Boolean = repository.existsByCompany(company)
    override fun existsByNameAndCompany(name: String, company: Company) =
        repository.existsByNameAndCompany(name, company)

    override fun isApprobationExpired(recipe: Recipe): Boolean? =
        with(Period.parse(configService.getContent(ConfigurationType.RECIPE_APPROBATION_EXPIRATION))) {
            recipe.approbationDate?.plus(this)?.isBefore(LocalDate.now())
        }

    override fun getAllByName(name: String) = repository.findAllByName(name)
    override fun getAllByCompany(company: Company) = repository.findAllByCompany(company)

    override fun save(entity: RecipeSaveDto): Recipe {
        val company = companyService.getById(entity.companyId)

        if (existsByNameAndCompany(entity.name, company)) {
            throw recipeNameAlreadyExistsForCompanyException(entity.name, company)
        }

        return save(with(entity) {
            recipe(
                name = name,
                description = description,
                color = color,
                gloss = gloss,
                sample = sample,
                approbationDate = approbationDate,
                remark = remark ?: "",
                company = company
            )
        })
    }

    @Transactional
    override fun update(entity: RecipeUpdateDto): Recipe {
        val persistedRecipe = getById(entity.id)
        val name = entity.name
        val company = persistedRecipe.company

        if (name != null &&
            name != persistedRecipe.name &&
            existsByNameAndCompany(name, company)
        ) {
            throw recipeNameAlreadyExistsForCompanyException(name, company)
        }

        return update(with(entity) {
            recipe(
                id = id,
                name = name or persistedRecipe.name,
                description = description or persistedRecipe.description,
                color = color or persistedRecipe.color,
                gloss = gloss ?: persistedRecipe.gloss,
                sample = sample ?: persistedRecipe.sample,
                approbationDate = approbationDate ?: persistedRecipe.approbationDate,
                remark = remark or persistedRecipe.remark,
                company = company,
                mixes = persistedRecipe.mixes,
                groupsInformation = updateGroupsInformation(persistedRecipe, entity)
            )
        })
    }

    private fun updateGroupsInformation(recipe: Recipe, updateDto: RecipeUpdateDto): Set<RecipeGroupInformation> {
        val steps = updateDto.steps ?: return recipe.groupsInformation

        val updatedGroupsInformation = mutableSetOf<RecipeGroupInformation>()
        steps.forEach {
            with(recipe.groupInformationForGroup(it.groupId)) {
                // Set steps for the existing RecipeGroupInformation or create a new one
                val updatedGroupInformation = this?.apply {
                    if (this.steps != null) {
                        this.steps!!.setAll(it.steps)
                    } else {
                        this.steps = it.steps.toMutableSet()
                    }
                } ?: recipeGroupInformation(
                    group = groupService.getById(it.groupId),
                    steps = it.steps.toMutableSet()
                )

                updatedGroupsInformation.add(updatedGroupInformation)
                recipeStepService.validateGroupInformationSteps(updatedGroupInformation)
            }
        }

        return updatedGroupsInformation
    }

    @Transactional
    override fun updatePublicData(publicDataDto: RecipePublicDataDto) {
        if (publicDataDto.notes != null) {
            val recipe = getById(publicDataDto.recipeId)

            fun noteForGroup(group: Group) =
                publicDataDto.notes.firstOrNull { it.groupId == group.id }?.content

            // Notes
            recipe.groupsInformation.map {
                val updatedNote = noteForGroup(it.group)
                it.apply {
                    note = updatedNote
                }
            }

            update(recipe)
        }

        if (publicDataDto.mixesLocation != null) {
            mixService.updateLocations(publicDataDto.mixesLocation)
        }
    }

    override fun addMix(recipe: Recipe, mix: Mix) =
        update(recipe.apply { mixes.add(mix) })

    override fun removeMix(mix: Mix): Recipe =
        update(mix.recipe.apply { mixes.remove(mix) })
}

interface RecipeImageService {
    /** Gets the name of every images associated to the recipe with the given [recipe]. */
    fun getAllImages(recipe: Recipe): Set<String>

    /** Saves the given [image] and associate it to the recipe with the given [recipe]. Returns the name of the saved image. */
    fun download(image: MultipartFile, recipe: Recipe): String

    /** Deletes the image with the given [name] for the given [recipe]. */
    fun delete(recipe: Recipe, name: String)

    /** Gets the directory containing all images of the given [Recipe]. */
    fun Recipe.getDirectory(): File
}

const val RECIPE_IMAGE_ID_DELIMITER = "_"
const val RECIPE_IMAGE_EXTENSION = ".jpg"

@Service
@Profile("!emergency")
class RecipeImageServiceImpl(
    val fileService: WriteableFileService
) : RecipeImageService {
    override fun getAllImages(recipe: Recipe): Set<String> {
        val recipeDirectory = recipe.getDirectory()
        if (!recipeDirectory.exists() || !recipeDirectory.isDirectory) {
            return setOf()
        }
        return recipeDirectory.listFiles()!!   // Should never be null because we check if recipeDirectory exists and is a directory before
            .filterNotNull()
            .map { it.name }
            .toSet()
    }

    override fun download(image: MultipartFile, recipe: Recipe): String {
        /** Gets the next id available for a new image for the given [recipe]. */
        fun getNextAvailableId(): Long =
            with(getAllImages(recipe)) {
                if (isEmpty())
                    0
                else
                    maxOf {
                        it.split(RECIPE_IMAGE_ID_DELIMITER)
                            .last()
                            .replace(RECIPE_IMAGE_EXTENSION, "")
                            .toLong()
                    } + 1L
            }

        return getImageFileName(recipe, getNextAvailableId()).apply {
            fileService.write(image, getImagePath(recipe, this), true)
        }
    }

    override fun delete(recipe: Recipe, name: String) =
        fileService.delete(getImagePath(recipe, name))

    override fun Recipe.getDirectory(): File = File(with(fileService) {
        this@getDirectory.imagesDirectoryPath.fullPath().path
    })

    private fun getImageFileName(recipe: Recipe, id: Long) =
        "${recipe.name}$RECIPE_IMAGE_ID_DELIMITER$id"

    private fun getImagePath(recipe: Recipe, name: String) =
        "${recipe.imagesDirectoryPath}/$name$RECIPE_IMAGE_EXTENSION"
}
