package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.model.ConfigurationType
import dev.fyloz.colorrecipesexplorer.model.touchupkit.*
import dev.fyloz.colorrecipesexplorer.repository.TouchUpKitRepository
import dev.fyloz.colorrecipesexplorer.rest.TOUCH_UP_KIT_CONTROLLER_PATH
import dev.fyloz.colorrecipesexplorer.service.config.ConfigurationService
import dev.fyloz.colorrecipesexplorer.service.files.FileService
import dev.fyloz.colorrecipesexplorer.service.files.WriteableFileService
import dev.fyloz.colorrecipesexplorer.utils.*
import org.springframework.context.annotation.Profile
import org.springframework.core.io.ByteArrayResource
import org.springframework.core.io.Resource
import org.springframework.stereotype.Service
import java.time.LocalDate
import java.time.Period

private const val TOUCH_UP_KIT_FILES_PATH = "pdf/touchupkits"

const val TOUCH_UP_TEXT_FR = "KIT DE RETOUCHE"
const val TOUCH_UP_TEXT_EN = "TOUCH UP KIT"

interface TouchUpKitService :
    ExternalModelService<TouchUpKit, TouchUpKitSaveDto, TouchUpKitUpdateDto, TouchUpKitOutputDto, TouchUpKitRepository> {
    fun isExpired(touchUpKit: TouchUpKit): Boolean

    fun complete(id: Long)

    /** Generates and returns a [PdfDocument] for the given [job]. */
    fun generateJobPdf(job: String): PdfDocument

    /**
     * Generates and returns a [PdfDocument] for the given [job] as a [ByteArrayResource].
     *
     * If TOUCH_UP_KIT_CACHE_PDF is enabled and a file exists for the job, its content will be returned.
     * If caching is enabled but no file exists for the job, the generated ByteArrayResource will be cached on the disk.
     */
    fun generateJobPdfResource(job: String): Resource

    /** Writes the given [document] to the [FileService] if TOUCH_UP_KIT_CACHE_PDF is enabled. */
    fun String.cachePdfDocument(document: PdfDocument)
}

@Service
@Profile("!emergency")
class TouchUpKitServiceImpl(
    private val fileService: WriteableFileService,
    private val configService: ConfigurationService,
    touchUpKitRepository: TouchUpKitRepository
) : AbstractExternalModelService<TouchUpKit, TouchUpKitSaveDto, TouchUpKitUpdateDto, TouchUpKitOutputDto, TouchUpKitRepository>(
    touchUpKitRepository
), TouchUpKitService {
    private val cacheGeneratedFiles by lazy {
        configService.getContent(ConfigurationType.TOUCH_UP_KIT_CACHE_PDF) == true.toString()
    }

    override fun idNotFoundException(id: Long) = touchUpKitIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = touchUpKitIdAlreadyExistsException(id)

    override fun TouchUpKit.toOutput() = TouchUpKitOutputDto(
        this.id!!,
        this.project,
        this.buggy,
        this.company,
        this.quantity,
        this.shippingDate,
        this.completionDate != null,
        this.completionDate,
        isExpired(this),
        this.finish,
        this.material,
        this.content,
        this.pdfUrl()
    )

    override fun update(entity: TouchUpKitUpdateDto): TouchUpKit {
        val persistedKit by lazy { getById(entity.id) }

        return super.update(with(entity) {
            touchUpKit(
                id = id,
                project = project ?: persistedKit.project,
                buggy = buggy ?: persistedKit.buggy,
                company = company ?: persistedKit.company,
                quantity = quantity ?: persistedKit.quantity,
                shippingDate = shippingDate ?: persistedKit.shippingDate,
                completionDate = completionDate ?: persistedKit.completionDate,
                finish = finish ?: persistedKit.finish,
                material = material ?: persistedKit.material,
                content = content?.map { touchUpKitProduct(it) }?.toSet() ?: persistedKit.content
            )
        })
    }

    override fun isExpired(touchUpKit: TouchUpKit) =
        with(Period.parse(configService.getContent(ConfigurationType.TOUCH_UP_KIT_EXPIRATION))) {
            touchUpKit.completed && touchUpKit.completionDate!!.plus(this) < LocalDate.now()
        }

    override fun complete(id: Long) {
        update(touchUpKitUpdateDto(id = id, completionDate = LocalDate.now()))
    }

    override fun generateJobPdf(job: String) = pdf {
        container {
            centeredVertically = true
            drawContainerBottom = true
            text(TOUCH_UP_TEXT_FR) {
                bold = true
                fontSize = PDF_DEFAULT_FONT_SIZE + 12
            }
            text(TOUCH_UP_TEXT_EN) {
                bold = true
                fontSize = PDF_DEFAULT_FONT_SIZE + 12
            }
            text(job) {
                marginTop = 10f
            }
        }

        container(containers[0]) {
            drawContainerBottom = false
        }
    }

    override fun generateJobPdfResource(job: String): Resource {
        if (cacheGeneratedFiles) {
            with(job.pdfDocumentPath()) {
                if (fileService.exists(this)) {
                    return fileService.read(this)
                }
            }
        }

        return generateJobPdf(job).apply {
            job.cachePdfDocument(this)
        }.toByteArrayResource()
    }

    override fun String.cachePdfDocument(document: PdfDocument) {
        if (!cacheGeneratedFiles) return

        fileService.write(document.toByteArrayResource(), this.pdfDocumentPath(), true)
    }

    private fun String.pdfDocumentPath() =
        "$TOUCH_UP_KIT_FILES_PATH/$this.pdf"

    private fun TouchUpKit.pdfUrl() =
        "${configService.getContent(ConfigurationType.INSTANCE_URL)}$TOUCH_UP_KIT_CONTROLLER_PATH/pdf?job=$project"
}
