package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.service.config.CONFIGURATION_FORMATTED_LIST_DELIMITER
import dev.fyloz.colorrecipesexplorer.service.config.ConfigurationServiceImpl
import dev.fyloz.colorrecipesexplorer.service.config.ConfigurationSource
import dev.fyloz.colorrecipesexplorer.service.files.ResourceFileService
import dev.fyloz.colorrecipesexplorer.service.files.WriteableFileService
import dev.fyloz.colorrecipesexplorer.utils.encrypt
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.core.io.Resource
import org.springframework.web.multipart.MultipartFile
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class ConfigurationServiceTest {
    private val fileService = mockk<WriteableFileService>()
    private val resourceFileService = mockk<ResourceFileService>()
    private val configurationSource = mockk<ConfigurationSource>()
    private val securityProperties = mockk<CreSecurityProperties> {
        every { configSalt } returns "d32270943af7e1cc"
    }
    private val service = spyk(
        ConfigurationServiceImpl(
            fileService,
            resourceFileService,
            configurationSource,
            securityProperties,
            mockk()
        )
    )

    @AfterEach
    fun afterEach() {
        clearAllMocks()
    }

    // getAll()

    @Test
    fun `getAll() gets the Configuration of each ConfigurationType`() {
        every { service.get(any<ConfigurationType>()) } answers { throw ConfigurationNotSetException(this.args[0] as ConfigurationType) }

        service.getAll()

        verify {
            service.getAll()
            ConfigurationType.values().forEach {
                service.get(it)
            }
        }
        confirmVerified(service)
    }

    @Test
    fun `getAll() only returns set configurations`() {
        val unsetConfigurationTypes = listOf(
            ConfigurationType.INSTANCE_NAME,
            ConfigurationType.INSTANCE_LOGO_SET,
            ConfigurationType.INSTANCE_ICON_SET
        )

        every { service.get(match<ConfigurationType> { it in unsetConfigurationTypes }) } answers {
            throw ConfigurationNotSetException(this.firstArg() as ConfigurationType)
        }
        every { service.get(match<ConfigurationType> { it !in unsetConfigurationTypes }) } answers {
            val type = firstArg<ConfigurationType>()
            configuration(type, type.key)
        }

        val found = service.getAll()

        assertFalse {
            found.any {
                it.type in unsetConfigurationTypes
            }
        }

        verify {
            service.getAll()
            ConfigurationType.values().forEach {
                service.get(it)
            }
        }
        confirmVerified(service)
    }

    @Test
    fun `getAll() only includes configurations matching the formatted formatted key list`() {
        val configurationTypes = listOf(
            ConfigurationType.INSTANCE_NAME,
            ConfigurationType.INSTANCE_LOGO_SET,
            ConfigurationType.INSTANCE_ICON_SET
        )
        val formattedKeyList = configurationTypes
            .map { it.key }
            .reduce { acc, s -> "$acc$CONFIGURATION_FORMATTED_LIST_DELIMITER$s" }

        every { service.get(any<String>()) } answers {
            val key = firstArg<String>()
            configuration(key.toConfigurationType(), key)
        }

        val found = service.getAll(formattedKeyList)

        assertTrue {
            found.all { it.type in configurationTypes }
        }

        verify {
            service.getAll(formattedKeyList)
            configurationTypes.forEach {
                service.get(it.key)
            }
        }
        confirmVerified(service)
    }

    // get()

    @Test
    fun `get(key) calls get() with the ConfigurationType matching the given key`() {
        val type = ConfigurationType.INSTANCE_ICON_SET
        val key = type.key

        every { service.get(type) } answers {
            val type = firstArg<ConfigurationType>()
            configuration(type, type.key)
        }

        service.get(key)

        verify {
            service.get(key)
            service.get(type)
        }
        confirmVerified(service)
    }

    @Test
    fun `get(type) gets the configuration in the ConfigurationSource`() {
        val type = ConfigurationType.INSTANCE_ICON_SET
        val configuration = configuration(type = type)

        every { configurationSource.get(type) } returns configuration

        val found = service.get(type)

        assertEquals(configuration, found)
    }

    @Test
    fun `get(type) throws ConfigurationNotSetException when the given ConfigurationType has no set configuration`() {
        val type = ConfigurationType.INSTANCE_ICON_SET

        every { configurationSource.get(type) } returns null

        with(assertThrows<ConfigurationNotSetException> { service.get(type) }) {
            assertEquals(type, this.type)
        }

        verify {
            service.get(type)
            configurationSource.get(type)
        }
    }

    @Test
    fun `get(type) throws InvalidConfigurationKeyException when the given ConfigurationType is encryption salt`() {
        val type = ConfigurationType.GENERATED_ENCRYPTION_SALT

        assertThrows<InvalidConfigurationKeyException> { service.get(type) }
    }

    @Test
    fun `get(type) returns a SecureConfiguration when the given ConfigurationType is secure`() {
        val type = ConfigurationType.DATABASE_PASSWORD
        val configuration = configuration(
            type = type,
            content = "securepassword".encrypt(type.key, securityProperties.configSalt!!)
        )

        every { configurationSource.get(type) } returns configuration

        val found = service.get(type)

        assertTrue { found is SecureConfiguration }
    }

    @Test
    fun `getContent(type) returns configuration content`() {
        val type = ConfigurationType.INSTANCE_NAME
        val configuration = configuration(
            type = type,
            content = "content"
        )

        every { service.get(type) } returns configuration

        val found = service.getContent(type)

        assertEquals(configuration.content, found)
    }

    @Test
    fun `getContent(type) throws UnsupportedOperationException when configuration is secure`() {
        val type = ConfigurationType.DATABASE_PASSWORD
        val configuration = secureConfiguration(type)

        every { service.get(type) } returns configuration

        assertThrows<UnsupportedOperationException> { service.getContent(type) }
    }

    @Test
    fun `getSecure(type) returns decrypted configuration content`() {
        val type = ConfigurationType.DATABASE_PASSWORD
        val content = "securepassword"
        val configuration = configuration(
            type = type,
            content = content.encrypt(type.key, securityProperties.configSalt!!)
        )

        every { configurationSource.get(type) } returns configuration

        val found = service.getSecure(type)

        assertEquals(content, found)
    }

    @Test
    fun `getSecure(type) throws UnsupportedOperationException when configuration is not secure`() {
        val type = ConfigurationType.INSTANCE_NAME

        assertThrows<UnsupportedOperationException> { service.getSecure(type) }
    }

    private fun getConfiguredImageTest(
        configurationType: ConfigurationType,
        imageSet: Boolean,
        test: (Resource) -> Unit
    ) {
        val resource = mockk<Resource>()
        val configuration = configuration(configurationType, imageSet.toString())
        val imageService = if (imageSet) fileService else resourceFileService

        every { service.get(configurationType) } returns configuration
        every { imageService.read(any()) } returns resource

        test(resource)
    }

    @Test
    fun `getConfiguredIcon() gets icon from resources when INSTANCE_ICON_SET configuration is false`() {
        getConfiguredImageTest(ConfigurationType.INSTANCE_ICON_SET, false) { resource ->
            val found = service.getConfiguredIcon()

            assertEquals(resource, found)
        }
    }

    @Test
    fun `getConfiguredIcon() gets icon from files when INSTANCE_ICON_SET configuration is true`() {
        getConfiguredImageTest(ConfigurationType.INSTANCE_ICON_SET, true) { resource ->
            val found = service.getConfiguredIcon()

            assertEquals(resource, found)
        }
    }

    @Test
    fun `getConfiguredLogo() gets logo from resources when INSTANCE_LOGO_SET is false`() {
        getConfiguredImageTest(ConfigurationType.INSTANCE_LOGO_SET, false) { resource ->
            val found = service.getConfiguredLogo()

            assertEquals(resource, found)
        }
    }

    @Test
    fun `getConfiguredLogo() gets logo from files when INSTANCE_LOGO_SET is true`() {
        getConfiguredImageTest(ConfigurationType.INSTANCE_LOGO_SET, true) { resource ->
            val found = service.getConfiguredLogo()

            assertEquals(resource, found)
        }
    }

    @Test
    fun `set(configuration) set configuration in ConfigurationSource`() {
        val configuration = configuration(type = ConfigurationType.INSTANCE_NAME)

        every { configurationSource.set(any<Configuration>()) } just runs

        service.set(configuration)

        verify {
            configurationSource.set(configuration)
        }
    }

    @Test
    fun `set(configuration) encrypts secure configurations`() {
        val type = ConfigurationType.DATABASE_PASSWORD
        val content = "securepassword"
        val encryptedContent = content.encrypt(type.key, securityProperties.configSalt!!)
        val configuration = configuration(type = type, content = content)

        mockkStatic(String::encrypt)

        every { configurationSource.set(any<Configuration>()) } just runs
        every { content.encrypt(any(), any()) } returns encryptedContent

        service.set(configuration)

        verify {
            configurationSource.set(match<Configuration> {
                it.content == encryptedContent
            })
        }
    }

    private fun setConfiguredImageTest(test: (MultipartFile) -> Unit) {
        val file = mockk<MultipartFile>()

        every { service.set(any<Configuration>()) } just runs
        every { fileService.write(any<MultipartFile>(), any(), any()) } just runs

        test(file)
    }

    @Test
    fun `setConfiguredIcon() sets icon in files`() {
        setConfiguredImageTest { file ->
            service.setConfiguredIcon(file)

            verify {
                fileService.write(file, any(), true)
            }
        }
    }

    @Test
    fun `setConfiguredIcon() sets INSTANCE_ICON_SET configuration to true`() {
        val type = ConfigurationType.INSTANCE_ICON_SET

        setConfiguredImageTest { file ->
            service.setConfiguredIcon(file)

            verify {
                service.set(match<Configuration> {
                    it.key == type.key && it.content == true.toString()
                })
            }
        }
    }

    @Test
    fun `setConfiguredLogo() sets logo in files`() {
        setConfiguredImageTest { file ->
            service.setConfiguredLogo(file)

            verify {
                fileService.write(file, any(), true)
            }
        }
    }

    @Test
    fun `setConfiguredLogo() sets INSTANCE_LOGO_SET configuration to true`() {
        val type = ConfigurationType.INSTANCE_LOGO_SET

        setConfiguredImageTest { file ->
            service.setConfiguredLogo(file)

            verify {
                service.set(match<Configuration> {
                    it.key == type.key && it.content == true.toString()
                })
            }
        }
    }
}
